from __future__ import annotations

import os
import subprocess

from devenv import main


def test(tmp_path: str) -> None:
    config_path = f"{tmp_path}/.config/sentry-devenv/config.ini"
    coderoot = f"{tmp_path}/code"
    reporoot = f"{coderoot}/repo"

    subprocess.run(("git", "init", reporoot))

    os.makedirs(f"{reporoot}/devenv")
    with open(f"{reporoot}/devenv/config.ini", "w") as f:
        pass
    with open(f"{reporoot}/devenv/sync.py", "w") as f:
        f.write(
            """
import devenv

def main(context: dict[str, str]) -> int:
    return 123
"""
        )

    # devenv sync should be able to be run from places other than reporoot...
    os.chdir(f"{reporoot}/devenv")

    rc = main.devenv(("/path/to/argv0", "sync"), config_path)
    assert rc == 123

    # ...and leave the cwd unchanged
    assert os.getcwd() == f"{reporoot}/devenv"



from __future__ import annotations

import os
import pathlib
from collections.abc import Iterator


def sorted_os_walk(
    path: pathlib.Path,
) -> Iterator[tuple[str, list[str], list[str]]]:
    for a, b, c in sorted(os.walk(path)):
        b.sort()
        c.sort()
        yield a, b, c



from __future__ import annotations

import os
import shutil
import tempfile

from devenv.lib import archive
from devenv.lib import fs
from devenv.lib import proc


def _install(url: str, sha256: str, into: str) -> None:
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")
        archive.unpack(archive_file, tmpd, perform_strip1=True)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/uv", f"{into}/uv")
        os.replace(f"{tmpd}/uvx", f"{into}/uvx")


def uninstall(binroot: str) -> None:
    for fp in (f"{binroot}/uv", f"{binroot}/uvx"):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def _version(binpath: str) -> str:
    stdout = proc.run((binpath, "--version"), stdout=True)
    # uv 0.7.21 (77c771c7f 2025-07-14)
    return stdout.split()[1]


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    binroot = fs.ensure_binroot(reporoot)
    binpath = f"{binroot}/uv"

    if shutil.which("uv", path=binroot) == binpath:
        installed_version = _version(binpath)
        if version == installed_version:
            return
        print(f"installed uv {installed_version} is unexpected!")

    print(f"installing uv {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    installed_version = _version(binpath)
    if version != installed_version:
        raise SystemExit("Failed to install uv {version}!")



from __future__ import annotations

import io
import os
import pathlib
import tarfile
import time
import typing
import urllib.request
from unittest import mock

import pytest

from devenv.lib import archive
from tests.utils import sorted_os_walk


@pytest.fixture
def tar(tmp_path: pathlib.Path) -> pathlib.Path:
    executable = tmp_path / "executable"
    executable.write_text("hi")

    tar = tmp_path / "tar"
    with tarfile.open(tar, "w:") as tarf:
        tarf.add(executable, arcname="executable")

    return tar


@pytest.fixture
def tgz(tmp_path: pathlib.Path) -> pathlib.Path:
    foo_v1 = tmp_path / "foo-v1"
    foo_v1.mkdir()

    foo_v1_bin = tmp_path / "foo-v1/bin"
    foo_v1_bin.mkdir()

    foo_v1_bin_foo = tmp_path / "foo-v1/bin/foo"
    foo_v1_bin_foo.write_text("")

    foo_v1_baz = tmp_path / "foo-v1/baz"
    foo_v1_baz.write_text("")

    tgz = tmp_path / "tgz"
    with tarfile.open(tgz, "w:gz") as tarf:
        tarf.add(foo_v1, arcname="foo-v1")
        # foo-v1
        # foo-v1/bin
        # foo-v1/bin/foo
        # foo-v1/baz

    return tgz


@pytest.fixture
def tar2(tmp_path: pathlib.Path) -> pathlib.Path:
    foo = tmp_path / "foo"
    foo.mkdir()

    foo_v1 = tmp_path / "foo/v1"
    foo_v1.mkdir()

    foo_v1_bin = tmp_path / "foo/v1/bin"
    foo_v1_bin.mkdir()

    foo_v1_bin_foo = tmp_path / "foo/v1/bin/foo"
    foo_v1_bin_foo.write_text("")

    foo_v1_baz = tmp_path / "foo/v1/baz"
    foo_v1_baz.mkdir()

    tar = tmp_path / "tar"
    with tarfile.open(tar, "w:") as tarf:
        tarf.add(foo, arcname="foo")
        # foo
        # foo/v1
        # foo/v1/bin
        # foo/v1/bin/foo
        # foo/v1/baz

    return tar


@pytest.fixture
def tar3(tmp_path: pathlib.Path) -> pathlib.Path:
    foo = tmp_path / "foo"
    foo.mkdir()

    foo_bar = tmp_path / "foo/bar"
    foo_bar.write_text("")

    foo_v1 = tmp_path / "foo/v1"
    foo_v1.mkdir()

    foo_v1_foo = tmp_path / "foo/v1/foo"
    foo_v1_foo.write_text("")

    tar = tmp_path / "tar"
    with tarfile.open(tar, "w:") as tarf:
        tarf.add(foo, arcname="foo")
        # foo
        # foo/bar
        # foo/v1
        # foo/v1/foo

    return tar


@pytest.fixture
def tar4(tmp_path: pathlib.Path) -> pathlib.Path:
    foo = tmp_path / "foo"
    foo.mkdir()

    foo_bar = tmp_path / "foo/bar"
    foo_bar.write_text("")

    tar = tmp_path / "tar"

    with tarfile.open(tar, "w:") as tarf:
        # note: arcname /foo doesn't work, it gets added as foo
        tarf.add(foo, arcname="foo")
        # /foo
        # /foo/bar
        for member in tarf.getmembers():
            member.path = f"/{member.path}"

    return tar


@pytest.fixture
def tar5(tmp_path: pathlib.Path) -> pathlib.Path:
    foo = tmp_path / "foo"
    foo.mkdir()

    foo_bar = tmp_path / "foo/bar"
    foo_bar.write_text("")

    foo_baz = tmp_path / "foo/baz"
    foo_baz.hardlink_to(foo_bar)

    tar = tmp_path / "tar"

    with tarfile.open(tar, "w:") as tarf:
        tarf.add(foo, arcname="foo")
        # foo
        # foo/bar
        # foo/baz -> foo/bar

    return tar


@pytest.fixture
def mock_sleep() -> typing.Generator[mock.MagicMock, None, None]:
    with mock.patch.object(time, "sleep", autospec=True) as mock_sleep:
        yield mock_sleep


def test_download(tmp_path: pathlib.Path, mock_sleep: mock.MagicMock) -> None:
    data = b"foo\n"
    data_sha256 = (
        "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c"
    )

    err = urllib.error.HTTPError(
        "https://example.com/foo",
        503,
        "Service Unavailable",
        "",  # type: ignore
        io.BytesIO(b""),
    )

    dest = f"{tmp_path}/a"

    # if dest is already a valid symlink it should be paved over
    with open(f"{tmp_path}/hi", "wb"):
        pass
    os.symlink(f"{tmp_path}/hi", dest)

    with mock.patch.object(
        urllib.request,
        "urlopen",
        autospec=True,
        side_effect=(err, err, err, io.BytesIO(data)),
    ):
        archive.download("https://example.com/foo", data_sha256, dest)

    # successful download after 3 retries
    assert mock_sleep.mock_calls == [
        mock.call(1.0),
        mock.call(2.0),
        mock.call(4.0),
    ]
    with open(dest, "rb") as f:
        assert f.read() == data


def test_download_exceeded_retries(
    tmp_path: pathlib.Path, mock_sleep: mock.MagicMock
) -> None:
    data_sha256 = (
        "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c"
    )

    err = urllib.error.HTTPError(
        "https://example.com/foo",
        503,
        "Service Unavailable",
        "",  # type: ignore
        io.BytesIO(b""),
    )
    dest = f"{tmp_path}/a"
    with pytest.raises(RuntimeError) as excinfo:
        with mock.patch.object(
            urllib.request,
            "urlopen",
            autospec=True,
            side_effect=(err, err, err, err),
        ):
            archive.download("https://example.com/foo", data_sha256, dest)

    # exceeded retry limits
    assert (
        f"{excinfo.value}"
        == "Error getting https://example.com/foo: HTTP Error 503: Service Unavailable"
    )


def test_download_dest_is_broken_symlink(
    tmp_path: pathlib.Path, mock_sleep: mock.MagicMock
) -> None:
    data = b"foo\n"
    data_sha256 = (
        "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c"
    )

    dest = f"{tmp_path}/a"

    # if dest is already a dead symlink it should be paved over as well
    os.symlink(f"{tmp_path}/does-not-exist", dest)

    with mock.patch.object(
        urllib.request,
        "urlopen",
        autospec=True,
        side_effect=(io.BytesIO(data),),
    ):
        archive.download("https://example.com/foo", data_sha256, dest)

    with open(dest, "rb") as f:
        assert f.read() == data


def test_download_wrong_sha(
    tmp_path: pathlib.Path, mock_sleep: mock.MagicMock
) -> None:
    data = b"foo\n"
    data_sha256 = (
        "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f4850b878ae4944c"
    )

    dest = f"{tmp_path}/a"

    with pytest.raises(RuntimeError) as excinfo:
        with mock.patch.object(
            urllib.request,
            "urlopen",
            autospec=True,
            side_effect=(io.BytesIO(data),),
        ):
            archive.download("https://example.com/foo", "wrong sha", dest)

    # sha mismatch
    assert (
        f"{excinfo.value}"
        == f"checksum mismatch for https://example.com/foo:\n- got: {data_sha256}\n- expected: wrong sha\n"
    )


def test_unpack_tar(tar: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path / "dest"
    archive.unpack(str(tar), str(dest))
    assert (dest / "executable").read_text() == "hi"


def test_unpack_tgz_strip1(tgz: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path / "dest"
    archive.unpack(str(tgz), str(dest), perform_strip1=True)

    assert [*sorted_os_walk(dest)] == [
        # bin/foo
        # baz
        (f"{dest}", ["bin"], ["baz"]),
        (f"{dest}/bin", [], ["foo"]),
    ]

    dest2 = tmp_path / "dest2"
    archive.unpack(
        str(tgz), str(dest2), perform_strip1=True, strip1_new_prefix="node"
    )

    assert [*sorted_os_walk(dest2)] == [
        # node/bin/foo
        # node/baz
        (f"{dest2}", ["node"], []),
        (f"{dest2}/node", ["bin"], ["baz"]),
        (f"{dest2}/node/bin", [], ["foo"]),
    ]


def test_unpack_strip_n(tar2: pathlib.Path, tmp_path: pathlib.Path) -> None:
    dest = tmp_path / "dest"
    archive.unpack_strip_n(str(tar2), str(dest), n=2)
    assert [*sorted_os_walk(dest)] == [
        # baz
        # bin/foo
        (f"{dest}", ["baz", "bin"], []),
        (f"{dest}/baz", [], []),
        (f"{dest}/bin", [], ["foo"]),
    ]

    dest2 = tmp_path / "dest2"
    archive.unpack_strip_n(str(tar2), str(dest2), n=2, new_prefix="x")
    assert [*sorted_os_walk(dest2)] == [
        # x/baz
        # x/bin/foo
        (f"{dest2}", ["x"], []),
        (f"{dest2}/x", ["baz", "bin"], []),
        (f"{dest2}/x/baz", [], []),
        (f"{dest2}/x/bin", [], ["foo"]),
    ]


def test_unpack_strip_n_unconditionally_removed(
    tar3: pathlib.Path, tmp_path: pathlib.Path
) -> None:
    dest = tmp_path / "dest"
    archive.unpack_strip_n(str(tar3), str(dest), n=2)

    # foo/bar is unconditionally removed

    assert [*sorted_os_walk(dest)] == [
        # dest/foo
        (f"{dest}", [], ["foo"])
    ]


def test_unpack_strip_n_root(
    tar4: pathlib.Path, tmp_path: pathlib.Path
) -> None:
    dest = tmp_path / "dest"
    archive.unpack_strip_n(str(tar4), str(dest), n=1)
    # leading slash in /foo/bar doesn't count as a component

    assert [*sorted_os_walk(dest)] == [
        # bar
        (f"{dest}", [], ["bar"])
    ]

    dest2 = tmp_path / "dest2"
    archive.unpack_strip_n(str(tar4), str(dest2), n=0)
    # n=0 can be used to just strip the root component

    assert [*sorted_os_walk(dest2)] == [
        # foo/bar
        (f"{dest2}", ["foo"], []),
        (f"{dest2}/foo", [], ["bar"]),
    ]


def test_unpack_strip_n_symlink(
    tar5: pathlib.Path, tmp_path: pathlib.Path
) -> None:
    dest = tmp_path / "dest"
    archive.unpack_strip_n(str(tar5), str(dest), n=1)
    # leading slash in /foo/bar doesn't count as a component

    assert [*sorted_os_walk(dest)] == [
        # bar
        # baz (-> bar)
        (f"{dest}", [], ["bar", "baz"])
    ]



from __future__ import annotations

import os
import shutil
import socket
import subprocess
import tempfile
from threading import Thread

from devenv.constants import home
from devenv.constants import root
from devenv.constants import SYSTEM_MACHINE
from devenv.lib import archive
from devenv.lib import proc


def _accept_and_close(sock: socket.socket) -> None:
    sock.listen()
    conn, addr = sock.accept()
    conn.close()


def check_docker_to_host_connectivity(timeout: int = 3) -> bool:
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]

    listener = Thread(target=_accept_and_close, args=(sock,))
    listener.start()

    rc = subprocess.call(
        (
            "docker",
            "run",
            "--rm",
            "--add-host=host.docker.internal:host-gateway",
            "busybox:1.36.1-musl",
            "/bin/sh",
            "-c",
            f"/bin/echo hi | /bin/nc -w {timeout} host.docker.internal {port}",
        )
    )

    if rc != 0:
        # easiest way to terminate the socket server
        # (so the thread doesn't indefinitely hang)
        with socket.socket() as s:
            s.connect(("127.0.0.1", port))
            s.send(b"die")

        listener.join()
        return False

    listener.join()
    return True


def uninstall(binroot: str) -> None:
    for fp in (
        f"{binroot}/docker",
        f"{home}/.docker/cli-plugins/docker-buildx",
    ):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")
        archive.unpack_strip_n(archive_file, tmpd, n=1)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/docker", f"{into}/docker")


def _install_buildx(url: str, sha256: str, into: str) -> None:
    archive.download(url, sha256, dest=f"{into}/docker-buildx")
    os.chmod(f"{into}/docker-buildx", 0o775)


def _check_buildx(binroot: str, expected_version: str) -> bool:
    try:
        stdout = proc.run(
            (f"{binroot}/docker", "buildx", "version"), stdout=True
        )
    except RuntimeError as e:
        print(f"failed getting buildx version:\n\n{e}")
        return False

    # github.com/docker/buildx v0.22.0 Homebrew
    installed_version = stdout.strip().split()[1]

    if installed_version == expected_version:
        return True

    print(
        f"installed docker-buildx {installed_version} is unexpected! expected: {expected_version}"
    )
    return False


def install_global() -> None:
    version = "27.3.1"
    cfg = {
        "darwin_x86_64": f"https://download.docker.com/mac/static/stable/x86_64/docker-{version}.tgz",
        "darwin_x86_64_sha256": "1b621d4c9a57ff361811cf29754aafb0c28bc113c70011927af8d73c2c162186",
        "darwin_arm64": f"https://download.docker.com/mac/static/stable/aarch64/docker-{version}.tgz",
        "darwin_arm64_sha256": "9dae125282116146b06eb777c2125ddda6c0468c0b9ad6c72a82edbc6783a77b",
        "linux_x86_64": f"https://download.docker.com/linux/static/stable/x86_64/docker-{version}.tgz",
        "linux_x86_64_sha256": "9b4f6fe406e50f9085ee474c451e2bb5adb119a03591f467922d3b4e2ddf31d3",
    }

    version_buildx = "v0.22.0"
    cfg_buildx = {
        "darwin_x86_64": f"https://github.com/docker/buildx/releases/download/{version_buildx}/buildx-{version_buildx}.darwin-amd64",
        "darwin_x86_64_sha256": "5221ad6b8acd2283f8fbbeebc79ae4b657e83519ca1c1e4cfbb9405230b3d933",
        "darwin_arm64": f"https://github.com/docker/buildx/releases/download/{version_buildx}/buildx-{version_buildx}.darwin-arm64",
        "darwin_arm64_sha256": "5898c338abb1f673107bc087997dc3cb63b4ea66d304ce4223472f57bd8d616e",
        "linux_x86_64": f"https://github.com/docker/buildx/releases/download/{version_buildx}/buildx-{version_buildx}.linux-amd64",
        "linux_x86_64_sha256": "805195386fba0cea5a1487cf0d47da82a145ea0a792bd3fb477583e2dbcdcc2f",
    }

    binroot = f"{root}/bin"

    if shutil.which("docker", path=binroot) == f"{binroot}/docker":
        stdout = proc.run((f"{binroot}/docker", "--version"), stdout=True)
        installed_version = stdout.strip().split()[2][:-1]
        if version != installed_version:
            print(f"installed docker {installed_version} is unexpected!")
            print(f"installing docker (cli, not desktop) {version}...")
            uninstall(binroot)
            _install(
                cfg[SYSTEM_MACHINE], cfg[f"{SYSTEM_MACHINE}_sha256"], binroot
            )
    else:
        _install(cfg[SYSTEM_MACHINE], cfg[f"{SYSTEM_MACHINE}_sha256"], binroot)

    stdout = proc.run((f"{binroot}/docker", "--version"), stdout=True)
    if f"Docker version {version}" not in stdout:
        raise SystemExit(f"Failed to install docker {version}!\n\n{stdout}")

    if not _check_buildx(binroot, version_buildx):
        print(f"installing docker buildx {version_buildx}...")
        _install_buildx(
            cfg_buildx[SYSTEM_MACHINE],
            cfg_buildx[f"{SYSTEM_MACHINE}_sha256"],
            f"{home}/.docker/cli-plugins",
        )

        if not _check_buildx(binroot, version_buildx):
            raise SystemExit(
                f"Failed to install docker buildx {version_buildx}!"
            )



from __future__ import annotations

import shlex
import subprocess
from pathlib import Path
from typing import Literal
from typing import overload

from devenv import constants
from devenv.constants import home
from devenv.constants import homebrew_bin
from devenv.constants import root
from devenv.constants import shell_path
from devenv.constants import user_environ

base_path = f"{root}/bin:{homebrew_bin}:{user_environ['PATH']}"
base_env = {"PATH": base_path, "HOME": home, "SHELL": shell_path}


def quote(cmd: tuple[str, ...]) -> str:
    """convert a command to bash-compatible form"""
    return " ".join(shlex.quote(arg) for arg in cmd)


def xtrace(cmd: tuple[str, ...]) -> None:
    """Print a commandline, similar to how xtrace does."""

    teal = "\033[36m"
    reset = "\033[m"
    bold = "\033[1m"

    print(f"+ {teal}${reset} {bold}{quote(cmd)}{reset}")


@overload
def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    stdout: Literal[False] = False,
) -> None:
    ...


@overload
def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    stdout: Literal[True],
) -> str:
    ...


def run(
    cmd: tuple[str, ...],
    *,
    pathprepend: str = "",
    exit: bool = False,
    env: dict[str, str] | None = None,
    cwd: Path | str | None = None,
    # poorly named, should've been like capture_combined_output
    stdout: bool = False,
) -> str | None:
    _stdout = subprocess.PIPE if stdout else None
    _stderr = subprocess.STDOUT if stdout else None
    del stdout

    if env is None:
        env = {}
    env = {**constants.user_environ, **base_env, **env}

    if pathprepend:
        env["PATH"] = f"{pathprepend}:{env['PATH']}"

    if constants.DEBUG:
        xtrace(cmd)
    try:
        proc = subprocess.run(
            cmd, check=True, stdout=_stdout, stderr=_stderr, cwd=cwd, env=env
        )
    except FileNotFoundError as e:
        # This is reachable if the command isn't found.
        if exit:
            raise SystemExit(f"{e}") from None
        else:
            raise RuntimeError(f"{e}") from None
    except subprocess.CalledProcessError as e:
        detail = f"Command `{quote(e.cmd)}` failed! (code {e.returncode})"
        if _stdout:
            detail += f"""
combined out:
{"" if e.stdout is None else e.stdout.decode()}
"""
        if exit:
            raise SystemExit(detail) from None
        else:
            raise RuntimeError(detail) from None
    else:
        if _stdout:
            return proc.stdout.decode().strip()
        return None



from __future__ import annotations

import json
import os
import shutil

from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = {"builtin"}
name = "correct docker configuration"


@checker
def check() -> tuple[bool, str]:
    # When Docker Desktop is opened, it inserts "desktop" as the value
    # for credsStore. This requires docker-credential-desktop which isn't
    # generally installed. Most people don't need to login to docker anyways.
    with open(os.path.expanduser("~/.docker/config.json"), "rb") as f:
        config = json.load(f)

    store = config.get("credsStore", "")
    if store and not shutil.which(f"docker-credential-{store}"):
        return False, "credsStore requires nonexistent binary"

    # When docker-buildx is installed via brew, brew adds cliPluginsExtraDirs
    # which takes precedence over the default plugin path we rely on.
    # This ensures the devenv-managed global docker cli uses the default plugin path.
    if config.get("cliPluginsExtraDirs"):
        return (
            False,
            "cliPluginsExtraDirs exists, which overshadows the default plugin path",
        )

    # Ensure the current context is set to colima
    current_context = config.get("currentContext", "")
    if current_context != "colima":
        return (
            False,
            f"currentContext is '{current_context}', should be 'colima'",
        )

    return True, ""


@fixer
def fix() -> tuple[bool, str]:
    try:
        with open(os.path.expanduser("~/.docker/config.json"), "rb") as f:
            config = json.load(f)

        config.pop("credsStore", None)
        config.pop("cliPluginsExtraDirs", None)
        config["currentContext"] = "colima"

        with open(os.path.expanduser("~/.docker/config.json"), "w") as f:
            json.dump(config, f)
    except Exception as e:
        return False, f"{e}"

    return True, ""



from __future__ import annotations

import time

from devenv.lib import proc
from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = {"builtin"}
name = "docker desktop shouldn't be running"


def docker_desktop_is_running() -> bool:
    procs = proc.run(("/bin/ps", "-Ac", "-o", "comm"), stdout=True)
    return "Docker Desktop" in procs.split("\n")


@checker
def check() -> tuple[bool, str]:
    if docker_desktop_is_running():
        return (
            False,
            "Docker Desktop is running. We don't support it, and it conflicts with colima.",
        )

    return True, ""


@fixer
def fix() -> tuple[bool, str]:
    # regular pkill won't stop the Docker Desktop UI,
    # it'll just spin. the proper way to terminate it
    # without SIGKILL is to use osascript.
    print("Attempting to stop Docker Desktop.")
    try:
        proc.run(("osascript", "-e", 'quit app "Docker Desktop"'))
    except RuntimeError as e:
        return False, f"failed to quit Docker Desktop:\n{e}\n"

    # osascript doesn't wait to make sure it finishes quitting
    # so let's block for up to 5 secs
    for _ in range(10):
        time.sleep(0.5)
        if not docker_desktop_is_running():
            return True, ""

    return False, "Docker Desktop is taking too long to quit... try again?"



from __future__ import annotations

import contextlib
import os
import pathlib
import shutil
import subprocess

import pytest

from devenv.lib.fs import gitroot
from devenv.lib.fs import write_script


def test_gitroot(tmp_path: pathlib.Path) -> None:
    subprocess.run(("git", "init", f"{tmp_path}"))

    with contextlib.chdir(tmp_path):
        assert os.path.samefile(tmp_path, gitroot())
        assert os.path.isdir(f"{gitroot()}/.git")


def test_gitroot_cd(tmp_path: pathlib.Path) -> None:
    subprocess.run(("git", "init", f"{tmp_path}"))
    os.mkdir(f"{tmp_path}/foo")

    _gitroot = gitroot(cd=f"{tmp_path}/foo")
    assert os.path.samefile(f"{tmp_path}", _gitroot)
    assert os.path.isdir(f"{_gitroot}/.git")


def test_no_gitroot(tmp_path: pathlib.Path) -> None:
    with pytest.raises(RuntimeError):
        with contextlib.chdir(tmp_path):
            gitroot()


def test_write_script(tmp_path: pathlib.Path) -> None:
    binroot = f"{tmp_path}"
    shim = "shim"
    bad = "$(echo hi)"

    write_script(
        f"{binroot}/{shim}",
        """#!/bin/sh
export PATH="{binroot}/node-env/bin:${{PATH}}"
export NPM_CONFIG_PREFIX="{binroot}/node-env"
very={bad}
exec "{binroot}/node-env/bin/{shim}" "$@"
""",
        shell_escape={"binroot": binroot, "shim": shim, "bad": bad},
    )

    with open(f"{binroot}/{shim}") as f:
        assert shutil.which(shim, path=binroot) == f"{binroot}/{shim}"
        text = f.read()
        assert (
            text
            == f"""#!/bin/sh
export PATH="{binroot}/node-env/bin:${{PATH}}"
export NPM_CONFIG_PREFIX="{binroot}/node-env"
very='$(echo hi)'
exec "{binroot}/node-env/bin/{shim}" "$@"
"""
        )



from __future__ import annotations

import os
import platform
import shutil
import tempfile
from enum import Enum

from devenv.constants import home
from devenv.constants import root
from devenv.constants import SYSTEM_MACHINE
from devenv.lib import archive
from devenv.lib import docker
from devenv.lib import fs
from devenv.lib import proc
from devenv.lib import rosetta


ColimaStatus = Enum("ColimaStatus", ("UP", "DOWN", "UNHEALTHY"))


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive.download(url, sha256, dest=f"{tmpd}/colima")
        os.replace(f"{tmpd}/colima", f"{into}/colima-bin")
        os.chmod(f"{into}/colima-bin", 0o775)


def uninstall(binroot: str) -> None:
    for d in (f"{home}/.lima",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (f"{binroot}/colima",):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def install_shim(binroot: str) -> None:
    fs.write_script(
        f"{binroot}/colima",
        """#!/bin/sh
export COLIMA_HOME={home}/.colima

# needed to ensure COLIMA_HOME is what we want it to be
unset XDG_CONFIG_HOME

exec {binroot}/colima-bin "$@"
""",
        shell_escape={"binroot": binroot, "home": home},
    )


def install_global() -> None:
    version = "v0.8.1"
    cfg = {
        "darwin_x86_64": f"https://github.com/abiosoft/colima/releases/download/{version}/colima-Darwin-x86_64",
        "darwin_x86_64_sha256": "791330c62c60389f70e5e1c33a56c35502a9e36e544a418daea0273e539acbf4",
        "darwin_arm64": f"https://github.com/abiosoft/colima/releases/download/{version}/colima-Darwin-arm64",
        "darwin_arm64_sha256": "c266fcb272b39221ef6152d2093bb02a1ebadc26042233ad359e1ae52d5d5922",
        "linux_x86_64": f"https://github.com/abiosoft/colima/releases/download/{version}/colima-Linux-x86_64",
        "linux_x86_64_sha256": "f2d6664a79ff3aa35f0718aac2ba9f6b531772e1464f3b096c1ac2aab404943e",
    }

    binroot = f"{root}/bin"

    if shutil.which("colima", path=binroot) == f"{binroot}/colima":
        if not os.path.exists(f"{binroot}/colima-bin"):
            os.rename(f"{binroot}/colima", f"{binroot}/colima-bin")
            install_shim(binroot)

        stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
        installed_version = stdout.strip().split()[-1]
        if version == installed_version:
            return
        print(f"installed colima {installed_version} is unexpected!")

    print(f"installing colima {version}...")
    uninstall(binroot)
    _install(cfg[SYSTEM_MACHINE], cfg[f"{SYSTEM_MACHINE}_sha256"], binroot)
    install_shim(binroot)

    if SYSTEM_MACHINE == "darwin_x86_64":
        if not shutil.which("qemu"):
            print(
                "WARNING: you're on darwin_x86_64, but QEMU isn't installed. Run: `brew install qemu`."
            )

    stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
    if f"colima version {version}" not in stdout:
        raise SystemExit(f"Failed to install colima {version}! Found: {stdout}")


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    install_global()


def check() -> ColimaStatus:
    if not os.getenv("CI"):
        macos_version = platform.mac_ver()[0]
        macos_major_version = int(macos_version.split(".")[0])
        if macos_major_version < 14:
            raise SystemExit(
                f"macos >= 14 is required to use colima, found {macos_version}"
            )

    docker_executable = shutil.which("docker")
    if not docker_executable:
        raise SystemExit(
            "docker executable not found, you might want to run devenv sync"
        )

    colima_executable = shutil.which("colima")
    if not colima_executable:
        raise SystemExit(
            "colima executable not found, try running devenv update (twice if this is your first time doing this) to install"
        )

    if not os.path.exists(f"{home}/.colima/default/docker.sock"):
        return ColimaStatus.DOWN

    # if colima's up, we should be able to communicate with that docker socket
    # at the most basic level
    try:
        proc.run(
            (docker_executable, "--context=colima", "version"), stdout=True
        )
    except RuntimeError:
        return ColimaStatus.UNHEALTHY

    # https://github.com/abiosoft/colima/issues/949
    healthy = docker.check_docker_to_host_connectivity()
    if not healthy:
        return ColimaStatus.UNHEALTHY

    return ColimaStatus.UP


def start(restart: bool = False) -> ColimaStatus:
    status = check()

    if status == ColimaStatus.UP:
        if not restart:
            return ColimaStatus.UP
        proc.run(("colima", "stop"), pathprepend=f"{root}/bin")
    elif status == ColimaStatus.DOWN:
        pass
    elif status == ColimaStatus.UNHEALTHY:
        print("colima seems to be unhealthy, stopping it")
        proc.run(("colima", "stop"), pathprepend=f"{root}/bin")

    # colima start will only WARN if rosetta is unavailable and keep going without it,
    # so we need to ensure it's installed and running ourselves
    rosetta.ensure()

    cpus = os.cpu_count()
    if cpus is None:
        raise SystemExit("failed to determine cpu count")

    # SC_PAGE_SIZE is POSIX 2008
    # SC_PHYS_PAGES is a linux addition but also supported by more recent MacOS versions
    SC_PAGE_SIZE = os.sysconf("SC_PAGE_SIZE")
    SC_PHYS_PAGES = os.sysconf("SC_PHYS_PAGES")
    if SC_PAGE_SIZE == -1 or SC_PHYS_PAGES == -1:
        raise SystemExit("failed to determine memsize_bytes")
    memsize_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")

    args = ["--cpu", f"{cpus//2}", "--memory", f"{memsize_bytes//(2*1024**3)}"]
    if platform.machine() == "arm64":
        args = [*args, "--vm-type=vz", "--vz-rosetta", "--mount-type=virtiofs"]

    # removing all docker contexts to ensure only colima context is created
    shutil.rmtree(f"{home}/.docker/contexts", ignore_errors=True)

    proc.run(
        (
            # we share the "default" machine across repositories
            "colima",
            "start",
            "--verbose",
            # this effectively makes the vm's resolvectl status use:
            # DNS Servers: 8.8.8.8 1.1.1.1 192.168.5.2
            # https://lima-vm.io/docs/config/network/user/
            # 192.168.5.2 is the host, accessible from the vm
            # sometimes using only the host will result in dns breaking
            # for any number of reasons (public wifi that gives you some weird dns server,
            # tethering, vpn, what have you)
            "--dns",
            "8.8.8.8",
            "--dns",
            "1.1.1.1",
            # ideally we keep ~ ro, but currently the "default" vm
            # is shared across repositories, so for ease of use we'll let home rw
            f"--mount=/var/folders:w,/private/tmp/colima:w,{home}:w",
            *args,
        ),
        pathprepend=f"{root}/bin",
    )

    proc.run(("docker", "context", "use", "colima"))

    status = check()
    return status


def restart() -> ColimaStatus:
    status = start(restart=True)
    return status


def stop() -> ColimaStatus:
    proc.run(("colima", "stop"), pathprepend=f"{root}/bin")
    status = check()
    return status



from __future__ import annotations

import os.path
import pathlib
import shutil
from collections.abc import Generator
from unittest import mock

import pytest

from devenv.checks import dockerConfig


@pytest.fixture
def fake_config(tmp_path: pathlib.Path) -> Generator[pathlib.Path]:
    cfg = tmp_path.joinpath("config.json")

    def new_expanduser(s: str) -> str:
        assert s == "~/.docker/config.json"
        return str(cfg)

    with mock.patch.object(os.path, "expanduser", new_expanduser):
        yield cfg


def test_no_credsStore_ok(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"currentContext": "colima"}')
    assert dockerConfig.check() == (True, "")


def test_binary_ok(fake_config: pathlib.Path) -> None:
    fake_config.write_text(
        '{"credsStore": "example", "currentContext": "colima"}'
    )
    with mock.patch.object(shutil, "which", return_value="/fake/exe"):
        assert dockerConfig.check() == (True, "")


@pytest.mark.parametrize("name", ("desktop", "osxkeychain"))
def test_binary_missing(fake_config: pathlib.Path, name: str) -> None:
    fake_config.write_text(f'{{"credsStore": "{name}"}}')
    with mock.patch.object(shutil, "which", return_value=None):
        assert dockerConfig.check() == (
            False,
            "credsStore requires nonexistent binary",
        )


def test_fix_credsStore(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"credsStore": "bad"}')
    assert dockerConfig.fix() == (True, "")
    assert fake_config.read_text() == '{"currentContext": "colima"}'


def test_fix_cliPluginsExtraDirs(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"cliPluginsExtraDirs": ["foo/"]}')
    assert dockerConfig.fix() == (True, "")
    assert fake_config.read_text() == '{"currentContext": "colima"}'


def test_currentContext_missing(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"auths": {}}')
    assert dockerConfig.check() == (
        False,
        "currentContext is '', should be 'colima'",
    )


def test_currentContext_wrong(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"currentContext": "desktop"}')
    assert dockerConfig.check() == (
        False,
        "currentContext is 'desktop', should be 'colima'",
    )


def test_currentContext_correct(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"currentContext": "colima"}')
    assert dockerConfig.check() == (True, "")


def test_fix_currentContext_missing(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"auths": {}}')
    assert dockerConfig.fix() == (True, "")
    assert (
        fake_config.read_text() == '{"auths": {}, "currentContext": "colima"}'
    )


def test_fix_currentContext_wrong(fake_config: pathlib.Path) -> None:
    fake_config.write_text('{"currentContext": "docker-desktop"}')
    assert dockerConfig.fix() == (True, "")
    assert fake_config.read_text() == '{"currentContext": "colima"}'



from __future__ import annotations

import argparse
import os
import sys
from collections.abc import Sequence

from devenv import constants
from devenv.lib import brew
from devenv.lib import colima
from devenv.lib import direnv
from devenv.lib import docker
from devenv.lib import limactl
from devenv.lib import proc
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo

module_help = "Updates global devenv and tools."


def main(context: Context, argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("version", type=str, nargs="?")
    parser.add_argument(
        "--post-update", action="store_true", help="Internal, do not use."
    )

    args = parser.parse_args(argv)

    # This is so that people don't have to run update twice.
    if args.post_update:
        # Mirror this in bootstrap.py.
        print(
            f"""\
Updating global tools (at {constants.root}/bin).
"""
        )
        os.makedirs(f"{constants.root}/bin", exist_ok=True)
        brew.install()
        docker.install_global()
        direnv.install()
        colima.install_global()
        limactl.install_global()
        return 0

    is_global_devenv = sys.executable.startswith(
        f"{constants.root}/venv/bin/python"
    )

    global_devenv_exists = os.path.exists(f"{constants.root}/bin/devenv")

    # even if we aren't the global devenv, we want to update
    # the global devenv (but only if it exists) out of convenience
    if is_global_devenv or global_devenv_exists:
        if args.version is None:
            version = "sentry-devenv"
        else:
            version = f"sentry-devenv=={args.version}"

        proc.run(
            (f"{constants.root}/venv/bin/pip", "install", "-U", version),
            env={
                # better than cli flag (who knows if it'll break in the future)
                "PIP_DISABLE_PIP_VERSION_CHECK": "1"
            },
        )

        print(
            f"""\

Global devenv at {constants.root}/bin/devenv updated.
"""
        )

    proc.run((sys.executable, "-P", "-m", "devenv", "update", "--post-update"))

    if not is_global_devenv:
        print(
            """\
To update the local devenv, you'll want to run `devenv sync`.
"""
        )

    return 0


module_info = DevModuleInfo(
    action=main, name=__name__, command="update", help=module_help
)



from __future__ import annotations

import hashlib
import os
import secrets
import shutil
import tarfile
import tempfile
import time
import urllib.request
from collections.abc import Generator
from collections.abc import Sequence
from urllib.error import HTTPError

from devenv.constants import home


def atomic_replace(src: str, dest: str) -> None:
    if os.path.dirname(src) != os.path.dirname(dest):
        raise RuntimeError(
            f"cannot atomically move to dest {dest}; it needs to be in the same dir as {src}"
        )
    os.replace(src, dest)


def download(
    url: str,
    sha256: str,
    dest: str = "",
    retries: int = 3,
    retry_exp: float = 2.0,
) -> str:
    if retries < 0:
        raise ValueError("Retries cannot be negative")

    if not dest:
        cache_root = f"{home}/.cache/sentry-devenv"
        dest = f"{cache_root}/{sha256}"
        os.makedirs(cache_root, exist_ok=True)

    if os.path.islink(dest):
        # there are cases where dest can be an existing symlink
        # (docker desktop starts and puts symlinks into ~/.docker/cli-plugins)
        # such symlinks should be removed otherwise callers to download
        # usually try to chmod after and end up with PermissionError
        os.remove(dest)

    if os.path.exists(dest):
        return dest

    headers = {}
    if url.startswith("https://ghcr.io/v2/homebrew"):
        # downloading homebrew blobs requires auth
        # you can get an anonymous token from https://ghcr.io/token?service=ghcr.io&scope=repository%3Ahomebrew/core/go%3Apull
        # but there's also a special shortcut token QQ==
        # https://github.com/Homebrew/brew/blob/2184406bd8444e4de2626f5b0c749d4d08cb1aed/Library/Homebrew/brew.sh#L993
        headers["Authorization"] = "bearer QQ=="

    req = urllib.request.Request(url, headers=headers)

    retry_sleep = 1.0
    while retries >= 0:
        try:
            resp = urllib.request.urlopen(req)
            break
        except HTTPError as e:
            if retries == 0:
                raise RuntimeError(f"Error getting {url}: {e}")
            print(f"Error getting {url} ({retries} retries left): {e}")

        time.sleep(retry_sleep)
        retries -= 1
        retry_sleep *= retry_exp

    dest_dir = os.path.dirname(dest)
    os.makedirs(dest_dir, exist_ok=True)

    with tempfile.NamedTemporaryFile(delete=False, dir=dest_dir) as tmpf:
        shutil.copyfileobj(resp, tmpf)
        tmpf.seek(0)
        checksum = hashlib.sha256()
        buf = tmpf.read(4096)
        while buf:
            checksum.update(buf)
            buf = tmpf.read(4096)

        if not secrets.compare_digest(checksum.hexdigest(), sha256):
            raise RuntimeError(
                f"checksum mismatch for {url}:\n"
                f"- got: {checksum.hexdigest()}\n"
                f"- expected: {sha256}\n"
            )

        atomic_replace(tmpf.name, dest)

    return dest


# strips the leading component unconditionally (like GNU tar)
# (/ is always stripped and doesn't count)
# if there are conflicting filepaths after this, they'll error during unpack
def strip1(
    members: Sequence[tarfile.TarInfo],
) -> Generator[tarfile.TarInfo, None, None]:
    for member in members:
        i = member.path.find("/")
        if i == -1:
            continue
        elif i == 0:
            i = member.path[1:].find("/") + 1
            if i == 0:
                continue

        member.path = member.path[i + 1 :]  # noqa: E203

        # hardlink target (linkname) is relative to the root of the
        # archive and must also be stripped
        if member.islnk():
            member.linkname = member.linkname[i + 1 :]  # noqa: E203

        yield member


def unpack(
    path: str,
    into: str,
    perform_strip1: bool = False,
    strip1_new_prefix: str = "",
) -> None:
    os.makedirs(into, exist_ok=True)
    with tarfile.open(name=path, mode="r:*") as tarf:
        members = tarf.getmembers()
        if perform_strip1:
            members = [_ for _ in strip1(members)]

        if strip1_new_prefix:
            for member in members:
                member.path = f"{strip1_new_prefix}/{member.path}"

        tarf.extractall(into, members=members, filter="tar")


def unpack_strip_n(path: str, into: str, n: int, new_prefix: str = "") -> None:
    os.makedirs(into, exist_ok=True)
    with tarfile.open(name=path, mode="r:*") as tarf:
        members = tarf.getmembers()

        for _ in range(n):
            members = [_ for _ in strip1(members)]

        if new_prefix:
            for member in members:
                member.path = f"{new_prefix}/{member.path}"

        tarf.extractall(into, members=members, filter="tar")



from __future__ import annotations

import os
import shutil
import tempfile

from devenv.constants import home
from devenv.constants import root
from devenv.constants import SYSTEM_MACHINE
from devenv.lib import archive
from devenv.lib import proc


def uninstall(binroot: str) -> None:
    for d in (f"{home}/.lima",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (
        f"{binroot}/lima",
        f"{binroot}/limactl",
        f"{binroot}/lima-guestagent.Linux-aarch64",
        f"{binroot}/lima-guestagent.Linux-x86_64",
        f"{binroot}/templates/default.yaml",
    ):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive_file = archive.download(url, sha256, dest=f"{tmpd}/download")

        # the archive from homebrew has a lima/version prefix
        if url.startswith("https://ghcr.io/v2/homebrew"):
            archive.unpack_strip_n(archive_file, tmpd, n=2)
        else:
            archive.unpack(archive_file, tmpd)

        # the archive was atomically placed into tmpd so
        # these are on the same fs and can be atomically moved too
        os.replace(f"{tmpd}/bin/lima", f"{into}/lima")
        os.replace(f"{tmpd}/bin/limactl", f"{into}/limactl")
        os.replace(
            f"{tmpd}/share/lima/lima-guestagent.Linux-aarch64",
            f"{into}/lima-guestagent.Linux-aarch64",
        )
        os.replace(
            f"{tmpd}/share/lima/lima-guestagent.Linux-x86_64",
            f"{into}/lima-guestagent.Linux-x86_64",
        )
        os.makedirs(f"{into}/templates", exist_ok=True)
        os.replace(
            f"{tmpd}/share/lima/templates/default.yaml",
            f"{into}/templates/default.yaml",
        )


def install_global() -> None:
    version = "1.0.3"
    cfg = {
        # upstream github releases aren't built for macOS 14, so for now
        # we use homebrew binaries from https://formulae.brew.sh/api/formula/lima.json
        # unfortunately there's no way to view previous versions of a formula
        # so if we ever need to upgrade to a specific non-latest version we'll
        # have to build it ourselves
        # sonoma
        "darwin_x86_64": "https://ghcr.io/v2/homebrew/core/lima/blobs/sha256:f675abee28d0f10f335f7a04dc6ad3de12434c83c2f0f32c913061204c137a94",
        "darwin_x86_64_sha256": "f675abee28d0f10f335f7a04dc6ad3de12434c83c2f0f32c913061204c137a94",
        # arm64_sonoma
        "darwin_arm64": "https://ghcr.io/v2/homebrew/core/lima/blobs/sha256:8aeb0a3b7295f0c3e0c2a7a92a798a44397936e5bb732db825aee6da5e762d7a",
        "darwin_arm64_sha256": "8aeb0a3b7295f0c3e0c2a7a92a798a44397936e5bb732db825aee6da5e762d7a",
        # on linux we use github releases since most people are probably not using
        # linuxbrew and the go binary in homebrew links to linuxbrew's ld.so
        "linux_x86_64": f"https://github.com/lima-vm/lima/releases/download/v{version}/lima-{version}-Linux-x86_64.tar.gz",
        "linux_x86_64_sha256": "b109cac29569a4aacab01c588f922ea6c7e2ef06ce9260bbc4c382e475bc3b98",
    }

    binroot = f"{root}/bin"

    if (
        shutil.which("lima", path=binroot) == f"{binroot}/lima"
        and shutil.which("limactl", path=binroot) == f"{binroot}/limactl"
    ):
        stdout = proc.run((f"{binroot}/limactl", "--version"), stdout=True)
        installed_version = stdout.strip().split()[-1]
        if version == installed_version:
            return
        print(f"installed limactl {installed_version} is unexpected!")

    uninstall(binroot)
    _install(cfg[SYSTEM_MACHINE], cfg[f"{SYSTEM_MACHINE}_sha256"], binroot)

    stdout = proc.run((f"{binroot}/limactl", "--version"), stdout=True)
    if f"limactl version {version}" not in stdout:
        raise SystemExit(
            f"Failed to install limactl {version}! Found: {stdout}"
        )


def install(version: str, url: str, sha256: str, reporoot: str) -> None:
    install_global()



from __future__ import annotations

import argparse
import os
import shutil
from collections.abc import Sequence

from devenv import constants
from devenv.lib import brew
from devenv.lib import colima
from devenv.lib import direnv
from devenv.lib import docker
from devenv.lib import github
from devenv.lib import limactl
from devenv.lib import proc
from devenv.lib import rosetta
from devenv.lib.config import Config
from devenv.lib.config import initialize_config
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import ExitCode


def main(context: Context, argv: Sequence[str] | None = None) -> ExitCode:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--default-config",
        action="append",
        help="Provide a default config value. e.g., -d coderoot:path/to/root",
    )

    args = parser.parse_args(argv)

    configs = {
        k: v for k, v in [i.split(":", 1) for i in args.default_config or []]
    }

    if "coderoot" not in configs and "code_root" in context:
        configs["coderoot"] = context["code_root"]

    default_config: Config = {"devenv": configs}
    initialize_config(context["config_path"], default_config)

    if not constants.CI and shutil.which("xcrun"):
        # xcode-select --install will take a while,
        # and involves elevated permissions with a GUI,
        # so best to just let the user go through that separately then retrying,
        # rather than waiting for it.
        # There is a way to perform a headless install but it's more complex
        # (refer to how homebrew does it).
        try:
            _ = proc.run(("xcrun", "-f", "git"), stdout=True)
        except RuntimeError:
            return "Failed to find git. Run xcode-select --install, then re-run bootstrap when done."

    # even though this is called before colima starts,
    # better to try and potentially (although unlikely) fail earlier rather than later
    rosetta.ensure()

    github.add_to_known_hosts()

    if not constants.EXTERNAL_CONTRIBUTOR and not github.check_ssh_access(
        # silence the error the first time since it's expected to happen
        silent=True
    ):
        is_employee = (
            False
            if constants.CI
            else input("Are you a Sentry employee? (Y/n): ").lower()
            in {"y", "yes", ""}
        )
        if not constants.CI and not is_employee:
            print(
                "Please set the SENTRY_EXTERNAL_CONTRIBUTOR environment variable and re-run bootstrap."
            )
            return 1
        pubkey = github.generate_and_configure_ssh_keypair()
        if not constants.CI:
            input(
                f"""
Failed to authenticate with an ssh key to GitHub.
We've generated and configured one for you at ~/.ssh/sentry-github.
Visit https://github.com/settings/ssh/new and add the following Authentication key:

{pubkey}

Then, you need to go to https://github.com/settings/keys, find your key,
and click Configure SSO, for the getsentry organization.

When done, hit ENTER to continue.
"""
            )
        while not github.check_ssh_access():
            input(
                "Still failing to authenticate to GitHub. ENTER to retry, otherwise ^C to quit."
            )

    # Mirror this in bootstrap.py.
    print(
        f"""\
Updating global tools (at {constants.root}/bin).
"""
    )
    os.makedirs(f"{constants.root}/bin", exist_ok=True)
    brew.install()
    docker.install_global()
    direnv.install()
    colima.install_global()
    limactl.install_global()

    os.makedirs(context["code_root"], exist_ok=True)

    print(
        """
All done! Please close this terminal window and start a fresh one.

Afterward, start working on your project using the devenv fetch command
e.g., devenv fetch sentry or devenv fetch ops
"""
    )

    return 0


module_info = DevModuleInfo(
    action=main,
    name=__name__,
    command="bootstrap",
    help="Bootstraps the development environment.",
)



from __future__ import annotations

import argparse
from collections.abc import Sequence

from devenv.lib import colima
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo

module_help = "Colima convenience commands."


def main(context: Context, argv: Sequence[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command", choices=("start", "restart", "check", "stop")
    )

    args = parser.parse_args(argv)

    # TODO: in addition to returning 1 we should print colima logs
    #       (and/or send to sentry)

    if args.command == "start":
        status = colima.start()
        if status == colima.ColimaStatus.UNHEALTHY:
            # https://github.com/abiosoft/colima/issues/949
            print("colima seems unhealthy, we'll try restarting once")
            status = colima.restart()
            if status != colima.ColimaStatus.UP:
                return 1
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "restart":
        status = colima.restart()
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "check":
        status = colima.check()
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "stop":
        status = colima.stop()
        if status != colima.ColimaStatus.DOWN:
            return 1

    return 0


module_info = DevModuleInfo(
    action=main, name=__name__, command="colima", help=module_help
)



from __future__ import annotations

import sys

from devenv.lib import colima
from devenv.lib import proc
from devenv.lib_check.types import checker
from devenv.lib_check.types import fixer

tags: set[str] = {"builtin"}
name = "colima's DNS isn't working"


@checker
def check() -> tuple[bool, str]:
    # dns resolution can... stop working if colima's running
    # and wifi changes to some other network that gives macos some
    # weird nameservers

    status = colima.check()
    if status != colima.ColimaStatus.UP:
        return False, "Colima isn't running."

    try:
        proc.run(
            (
                "colima",
                "exec",
                "--",
                "python3",
                "-Su",
                "-c",
                """
import socket

try:
    socket.getaddrinfo("ghcr.io", None)
except socket.gaierror as e:
    raise SystemExit(f"failed to resolve ghcr.io: {e}")
""",
            ),
            stdout=True,
        )
    except RuntimeError as e:
        return False, f"{e}"

    return True, ""


@fixer
def fix() -> tuple[bool, str]:
    status = colima.start()
    if status == colima.ColimaStatus.UNHEALTHY:
        return False, "colima started, but it's unhealthy"

    try:
        proc.run(
            (
                "colima",
                "exec",
                "--",
                "sudo",
                "systemctl",
                "restart",
                "systemd-resolved.service",
            ),
            stdout=True,
        )
    except RuntimeError as e:
        print(
            f"""
failed to restart the vm's resolved:
{e}

we're going to try restarting colima
"""
        )
        try:
            proc.run(
                (sys.executable, "-P", "-m", "devenv", "colima", "restart")
            )
        except RuntimeError as e:
            return False, f"{e}"

    return True, ""



from __future__ import annotations

import os
from unittest.mock import patch

import pytest

from devenv.lib.repository import Repository


def test_check_minimum_version(tmp_path: str) -> None:
    repo = Repository(f"{tmp_path}/repo")

    mock_config = """
[devenv]
minimum_version = 1.11.0
"""

    os.makedirs(repo.config_path)
    with open(f"{repo.config_path}/config.ini", "w") as f:
        f.write(mock_config)

    with patch("importlib.metadata.version", return_value="1.10.2"):
        with pytest.raises(SystemExit):
            repo.check_minimum_version()

    with patch("importlib.metadata.version", return_value="1.11.0"):
        repo.check_minimum_version()

    with patch("importlib.metadata.version", return_value="1.11.1"):
        repo.check_minimum_version()
